#include "z166_kde4_ffmpeg_thumbnails.h"

FFmpeg::FFmpeg( QImage &img, int _w, int _h )
{
	chkKeyFrame = true;
	formatCtx = NULL;
	codecCtx = NULL;
	img_convert_ctx = NULL;
	buffer = NULL;
	videoStream = -1;
	W = H = 0;
	w = _w;
	h = _h;
	image = &img;
}
FFmpeg::~FFmpeg()
{
	if ( codecCtx )
		avcodec_close( codecCtx );
	if ( formatCtx )
		av_close_input_file( formatCtx );
	if ( img_convert_ctx )
		sws_freeContext( img_convert_ctx );
	if ( buffer )
		av_free( buffer );
}
bool FFmpeg::seekToFrame( uint64_t sec )
{
	bool OK = av_seek_frame( formatCtx, -1, AV_TIME_BASE * sec, 0 ) >= 0;
	avcodec_flush_buffers( formatCtx->streams[ videoStream ]->codec );
	return OK;
}
bool FFmpeg::decodeFrame( bool getImage = true )
{
	AVPacket packet;
	for (;;)
	{
		if ( av_read_frame( formatCtx, &packet ) )
			return false;
		if ( packet.stream_index == videoStream )
		{
			AVFrame frame;
			int frameFinished;
#if LIBAVCODEC_VERSION_MINOR < 37
			avcodec_decode_video( codecCtx, &frame, &frameFinished, packet.data, packet.size );
#else
			avcodec_decode_video2( codecCtx, &frame, &frameFinished, &packet );
#endif
			if ( frameFinished )
			{
				if ( !getImage )
				{
					if ( packet.data )
						av_free_packet( &packet );
					return true;
				}

				if ( chkKeyFrame && !frame.key_frame ) //sprawdzane tylko wtedy, gry był seek
				{
					if ( packet.data )
						av_free_packet( &packet );
					continue;
				}

				sws_scale( img_convert_ctx, frame.data, frame.linesize, 0, H, frameRGB.data, frameRGB.linesize );
				if ( packet.data )
					av_free_packet( &packet );
				*image = QImage( *frameRGB.data, w, h, QImage::Format_RGB32 ).copy();
				return true;
			}
		}
	}
	return false;
}
void FFmpeg::read( QString fName )
{
	uint firstXSeconds = 8;
	double seekTo = 0.4;
	bool readFirstFrame = false;
	QSettings settings( settingsFile, QSettings::IniFormat );
	if ( settings.contains( "z166_kde4_ffmpeg_thumbnails/blacklistedFileExtensions" ) )
	{
		QStringList tmp = fName.split( '.' );
		if ( tmp.count() > 1 )
		{
			QString fNameExt = tmp[ tmp.count() - 1 ].toLower();
			if ( settings.value( "z166_kde4_ffmpeg_thumbnails/blacklistedFileExtensions" ).toStringList().contains( fNameExt ) )
				return;
		}
	}
	if ( settings.contains( "z166_kde4_ffmpeg_thumbnails/readFirstXSecondsWhenCantSeek" ) )
	{
		firstXSeconds = settings.value( "z166_kde4_ffmpeg_thumbnails/readFirstXSecondsWhenCantSeek" ).toInt(); //default 8, max 10
		if ( firstXSeconds > 10 )
			firstXSeconds = 10;
	}
	if ( settings.contains( "z166_kde4_ffmpeg_thumbnails/lookForKeyframeWhenSeek" ) )
		chkKeyFrame = settings.value( "z166_kde4_ffmpeg_thumbnails/lookForKeyframeWhenSeek" ).toBool(); //default true
	if ( settings.contains( "z166_kde4_ffmpeg_thumbnails/seekTo" ) )
	{
		seekTo = settings.value( "z166_kde4_ffmpeg_thumbnails/seekTo" ).toDouble(); //default 0.4
		if ( seekTo < 0. || seekTo > 0.9 )
			seekTo = 0.4;
	}
	if ( settings.contains( "z166_kde4_ffmpeg_thumbnails/alwaysReadFirstFrame" ) )
		readFirstFrame = settings.value( "z166_kde4_ffmpeg_thumbnails/alwaysReadFirstFrame" ).toBool(); //default false

	if ( av_open_input_file( &formatCtx, fName.toUtf8(), NULL, 0, NULL ) )
		return;
	if ( av_find_stream_info( formatCtx ) < 0 )
		return;
	for ( uint i = 0 ; i < formatCtx->nb_streams ; i++ )
	{
		if ( formatCtx->streams[ i ]->codec->codec_type == CODEC_TYPE_VIDEO )
		{
			videoStream = i;
			break;
		}
	}
	if ( videoStream == -1 )
		return;
	codecCtx = formatCtx->streams[ videoStream ]->codec;
	AVCodec *codec = avcodec_find_decoder( codecCtx->codec_id );
	if ( avcodec_open( codecCtx, codec ) )
		return;

	W = codecCtx->width;
	H = codecCtx->height;

	double FPS = av_q2d( formatCtx->streams[ videoStream ]->r_frame_rate );
	uint64_t duration = formatCtx->duration / 1000000LL;

	double aspect_ratio = av_q2d( codecCtx->sample_aspect_ratio );
	if ( aspect_ratio )
		aspect_ratio *= ( (double)W / (double)H );
	else
		aspect_ratio = (double)W / (double)H;
	aspect_ratio /= (double)w / (double)h;
	if ( aspect_ratio > 1. )
		h /= aspect_ratio;
	else if ( aspect_ratio < 1. )
		w *= aspect_ratio;

	buffer = (uchar*)av_malloc( avpicture_get_size( PIX_FMT_RGB32, w, h ) );
	avpicture_fill( &frameRGB, buffer, PIX_FMT_RGB32, w, h );

	img_convert_ctx = sws_getContext( W, H, codecCtx->pix_fmt, w, h, PIX_FMT_RGB32, SWS_POINT, NULL, NULL, NULL );

	if ( !decodeFrame( readFirstFrame ) )
		return;
	int sek = duration * seekTo;
	bool seekOK = seekToFrame( sek );
	if ( !seekOK || ( seekOK && !decodeFrame() ) )
	{
		chkKeyFrame = false;
		seekToFrame( 0 );
		if ( duration > firstXSeconds )
			sek = firstXSeconds;
		for ( int i = 0 ; i < FPS * sek ; i++ )
			if ( !decodeFrame() )
				return;
	}
}

extern "C"
{
	KDE_EXPORT ThumbCreator *new_creator()
	{
		return new z166_kde4_ffmpeg_thumbnails;
	}
}

z166_kde4_ffmpeg_thumbnails::z166_kde4_ffmpeg_thumbnails()
{
	av_register_all();
}
bool z166_kde4_ffmpeg_thumbnails::create( const QString &fName, int w, int h, QImage &img )
{
	FFmpeg *ffmpeg = new FFmpeg( img, w, h );
	ffmpeg->read( fName );
	delete ffmpeg;
	return !img.isNull();
}
